(*****************************************************************)
(*                                                               *)
(*  ASDL to Caml translator                                      *)
(*  Copyright (c) 2000 by Dmitry Lomov (dsl@tepkom.ru)           *)
(*  Distributed under GNU Public License                         *)
(*                                                               *)
(*****************************************************************)

(* $Id: gen.ml,v 1.1.1.1 2000/12/05 09:06:52 dsl Exp $ *)

open AsdlDef
open Pp

let new_type_prefix = "_asdl_"

(* 
 * Type declarations 
 *)

let print_typeid (mod_name, type_name) =
  (match mod_name with
    Some id ->  format (id^".")
  | None -> ());
  format type_name

let print_field_type m t =
  print_typeid t;
  match m with
    Straight -> ()
  | Optional -> format " option"
  | Sequence -> format " list"   

let constructor_type_name id = new_type_prefix^id 

let attributes_type_name tid =
  let p = new_type_prefix^"attributes_" in
  p^tid

let print_field (_,id,m,typ) =
  format id; format " : "; print_field_type m typ
 
let print_named_record l =
  let rec print_field_list = function
      [] -> ()
    | [f] -> print_field f
    | h::t -> 
	print_field h; format "; "; print_field_list t in
  format "{"; print_field_list l; format "}"

let rec print_product = function
    [] -> ()
  | [(_,_,m,typ)] -> print_field_type m typ
  | (_,_,m,typ)::t -> 
      print_field_type m typ; format " * "; print_product t

let print_record l =
  if (List.for_all 
	(fun (mid,_,_,_) -> match mid with Generated -> true | _ -> false) l)
  then print_product l
  else print_named_record l
    

let rec print_constr_types = function
    [] -> ()
  | (id,fl)::t -> 
      (match fl with
	[] -> ()
      |	_ ->
	  format (constructor_type_name id); format " =";
	  print_record fl; format " and";
	  indent());
      print_constr_types t 

let rec print_attributes tid attrs =
  format (attributes_type_name tid); format " =";
  print_record attrs; format " and"; 
  indent()

let rec (print_constructors : id -> constructor list -> bool -> unit) =
  fun tid constr attr_present ->
  let rec print_constructor (uid, fl) =
    format uid; 
    match fl, attr_present with
      [],true ->
	format " of "; format (attributes_type_name tid)
    | [],false -> ()
    | _, false ->
	format " of "; format (constructor_type_name uid)
    | _, true ->
	format " of "; format (constructor_type_name uid);
	format " * "; format (attributes_type_name tid) 
  in
  match constr with
    [] -> ()
  | [c] -> 
      print_constructor c
  | c::t -> 
      print_constructor c; 
      indent(); format "| "; 
      print_constructors tid t attr_present

let print_type : typeDef -> unit =
  fun (tid,expr) ->
    match expr with
      Product l ->
	indent();
	format tid; format " ="; print_record l
    | Sum (constr, attr) ->
	indent();
	print_constr_types constr;
	(match attr with 
	  [] -> ()
	| _ -> print_attributes tid attr);
	format tid; format " =";
	print_constructors tid constr (match attr with [] -> false | _ -> true)

let rec print_type_list = function
    [] -> ()
  | [t] -> 
      print_type t 
  | t::rest ->
      print_type t;
      format " and";
      print_type_list rest

let gen_types typs = 
  indent (); format "type";
  incr ();
  print_type_list typs;
  decr ()

(*
 * I/O signature
 *)
let rec (gen_io_sig : typeDef list -> unit) = function
    [] -> ()
  | (id, _)::t ->
      indent ();
      format "val ";
      format ("write_"^id); format " : ";
      format id; format " -> out_channel -> unit";
      indent ();
      format "val ";
      format ("write_tagged_"^id); format " : ";
      format id; format " -> out_channel -> unit";
      indent ();
      format "val ";
      format ("read_"^id); format " : ";
      format "in_channel -> "; format id;
      indent ();
      format "val ";
      format ("read_tagged_"^id); format " : ";
      format "in_channel -> "; format id;
      gen_io_sig t

(*
 * I/O implementation
 *)

let gen_field_pattern is_first (_,id,_,(m_id,_)) =
  if not is_first then format "; ";
  format id; 
  format " = "; format id; 
  false 

let gen_anon_field_pattern is_first (_,id,_,_) =
  if not is_first then format ", ";
  format id;
  false
  

let gen_record_pattern : field list -> unit =
  fun fields ->
    let gen_pattern, c1, c2 = 
      if (List.for_all (fun (id_m,_,_,_) -> id_m = Generated) fields)
      then gen_anon_field_pattern, "( ", " )"
      else gen_field_pattern, "{ ", " }" in
    
    match fields with
      [] -> ()
    | _ -> 
	format c1; 
	ignore (List.fold_left gen_pattern true fields);
	format c2

(*=== Writing  ===*)
let gen_write_field is_first (_,id,m,(m_id,t_id)) =
  let writer_name = 
    (match m_id with
      None -> ""
    | Some name -> name^".")^"write_"^t_id in
  if not is_first 
  then (format ";"; indent ());
  (match m with
    Straight -> ()
  | Sequence -> format "write_list "
  | Optional -> format "write_option ");
  format writer_name; format (" "^id); format " _s";
  false

let gen_write_product id fields =
  format ("write_"^id); 
  gen_record_pattern fields;
  format "_s"; format " =";
  incr (); indent ();
  ignore (List.fold_left gen_write_field true fields);
  decr ()


let gen_write_sum (id : id) (constr : constructor list) (attr : field list) =
  let gen_write_constructor tag (id, fields) =
    if tag == 1
    then format "  "
    else format "| ";
    format id;
    (match fields,attr with
      [],_ | _,[] -> 
	();
    | _ -> 
	format "(");
    gen_record_pattern fields;
    (match attr,fields with
      [],[] -> ()
    | [], _ | _,[] -> gen_record_pattern attr
    | _ -> format ","; gen_record_pattern attr; format ")");
    format " ->";
    incr (); incr (); indent ();
    format "write_tag "; format (string_of_int tag); format " _s";
    format ";"; indent ();
    ignore(List.fold_left gen_write_field true attr);
    ignore(List.fold_left gen_write_field (attr==[]) fields);    
    decr (); decr (); indent (); tag + 1 in
  format ("write_"^id);  format " _x"; format " _s"; format " =";
  incr (); indent ();
  format "match "; format "_x "; format "with";
  indent ();
  ignore (List.fold_left gen_write_constructor 1 constr);
  decr ()


let gen_write_type id expr =
  match expr with
    Product fields -> gen_write_product id fields
  | Sum(c,a) -> gen_write_sum id c a 

(*=== Reading ===*)

let gen_read_field (_, id, mt, (mid, tid)) =
  indent ();
  format "let "; format id; format " = ";
  (match mt with
    Straight -> ()
  | Sequence -> format "read_list "
  | Optional -> format "read_option ");
  (match mid with
    None -> ()
  | Some name -> format name; format ".");
  format ("read_"^tid); format " _s"; format " in"

let gen_read_product id fields = 
  ignore (List.iter gen_read_field fields);
  indent ();
  gen_record_pattern fields;
  decr()
  
let gen_read_sum id constr attr =
  let gen_read_constructor tag (id, fields) =
    if tag == 1
    then format "  "
    else format "| ";
    format (string_of_int tag);
    format " ->";
    incr(); indent();
    ignore (List.iter gen_read_field attr);
    ignore (List.iter gen_read_field fields);
    indent ();
    format id;
    (match fields,attr with
      [], [] -> ()
    |  _, [] -> gen_record_pattern fields 
    | [], _  -> gen_record_pattern attr 
    | _ -> 
	format "(";
	gen_record_pattern fields;
	format ",";
	gen_record_pattern attr;
	format ")");
    decr ();
    indent (); tag + 1 in
  indent ();
  format "match "; format "read_tag "; format "_s "; format "with"; 
  indent ();
  ignore (List.fold_left gen_read_constructor 1 constr);
  format "| "; format "_ "; format "->"; format " die()"

let gen_read_type id expr =
  format ("read_"^id); format " _s"; format " =";
  incr ();
  (match expr with
    Product fields ->
      gen_read_product id fields
  | Sum(c,a) ->
      gen_read_sum id c a);
  decr ()

let gen_io_type tag (id,expr) =
  let stag = (string_of_int tag) in
  gen_write_type id expr;
  indent (); format "and ";
  format ("write_tagged_"^id); format " x"; format " s"; format " = ";
  incr (); indent ();
  format "("; format "write_tag "; 
  format stag; format " s";
  format "; "; format ("write_"^id); format " x"; format " s"; format ")";
  decr () ; indent (); 
  format "and "; gen_read_type id expr;
  indent (); format "and ";
  format ("read_tagged_"^id); format " s"; format " = ";
  incr (); indent ();
  format "if "; format "read_tag "; format "s "; format "== "; format stag;
  indent ();
  format "then "; format ("read_"^id); format " s";
  indent ();
  format "else "; format "die ()"; 
  decr ();
  (tag + 1)

let gen_io_struct next_tag (typ::ttyp) =
  indent ();
  format "let "; format "rec ";
  let next_tag = gen_io_type next_tag typ in
  List.fold_left
    (fun tag t -> indent (); format "and "; gen_io_type tag t)
    next_tag ttyp  
    
(*
 * genModule
 *)
let modules : (id,int) Hashtbl.t = Hashtbl.create 10
exception UndefinedModule of string

let check_imports imports = 
  let check_this tag id =
    try 
      max tag (Hashtbl.find modules id)
    with
      Not_found -> raise (UndefinedModule id) in
  List.fold_left check_this 4 imports

let genModule : moduleDef -> unit = 
  fun  def ->
    let (id,imports,typs) = Naming.naming def in
    try 
      let next_tag = check_imports imports in
      let next_tag =
	if not Params.v.Params.separate 
	then
	  (format "module "; format  id; 
	   format " : "; format "sig"; 
	   incr(); 
	   indent (); format "open "; format "AsdlBase"; indent();
	   gen_types typs;
	   gen_io_sig typs;
	   decr ();
	   format "end" ;
	   format " =";
	   incr ();
	   indent (); format "struct";
	   incr ();
	   indent (); format "open "; format "AsdlBase"; indent();
	   gen_types typs;
	   let next_tag = gen_io_struct next_tag typs in
	   decr();
	   indent (); format "end";
	   decr ();
	   indent ();
	   next_tag)
	else
	  let fn = String.uncapitalize id in
	  let fnin = fn^".mli" and fnim = fn^".ml" in 
	  let f = open_out fnin in
	  if Params.v.Params.filenames 
	  then (Printf.printf "%s\n" fnin; flush stdout);
	  set_channel f;
	  format "open "; format "AsdlBase"; indent();
	  gen_types typs;
	  gen_io_sig typs;
	  flush f;
	  let f = open_out fnim in
	  if Params.v.Params.filenames then Printf.printf "%s\n" fnim; 
	  set_channel f;
	  format "open "; format "AsdlBase"; indent();
	  gen_types typs;
	  let next_tag = gen_io_struct next_tag typs in
	  flush f;
	  next_tag in
      Hashtbl.add modules id next_tag
    with
      Sys_error s ->
	Params.msg Params.Error s None
    | UndefinedModule uid ->
	Params.msg Params.Error ("Undefined module: "^uid) None
	  
      

